from threading import Thread
from .client import Client
from .exceptions import AIHeroException
from .openai_helper import get_embedding, has_key
from warnings import warn


class PromptStash:
    def __init__(
        self,
        project_id: str,
        client: Client,
    ):
        self._project_id = project_id
        self._client = client

    def variant(
        self,
        template_id: str,
        variant: str = None,
        body: str = None,
        name: str = None,
        description: str = None,
        prompt_format: str = "f-string",
        sections: str = None,
        author: str = None,
        other: dict = None,
    ):
        if variant is None and body is None:
            prompt_template_dict = self._client.get(
                f"/tools/promptstash/projects/{self._project_id}/prompt_templates/{template_id}",
                error_msg=f"Could fetch prompt template {template_id} for project {self._project_id}",
                network_errors={
                    400: "Please check the prompt_template_id.",
                    403: f"Could not access project {self._project_id}. Please check the API key.",
                },
            )
        else:
            if variant:
                prompt_template_dict = self._client.get(
                    f"/tools/promptstash/projects/{self._project_id}/prompt_templates/{template_id}/variants/{variant}",
                    error_msg=f"Could fetch variant {variant} of prompt template {template_id} for project {self._project_id}",
                    network_errors={
                        400: "Please check the prompt_variant.",
                        403: f"Could not access project {self._project_id}. Please check the API key.",
                    },
                )
            else:
                prompt_template_dict = {
                    "body": body,
                    "prompt_format": prompt_format,
                    "sections": {},
                    "author": "",
                    "other": {},
                }
                if name:
                    prompt_template_dict["name"] = name
                if description:
                    prompt_template_dict["description"] = description
                if sections:
                    prompt_template_dict["sections"] = sections
                if author:
                    prompt_template_dict["author"] = author
                if other:
                    prompt_template_dict["other"] = other
                # No variant found
                prompt_template_dict = self._client.post(
                    f"/tools/promptstash/projects/{self._project_id}/prompt_templates/{template_id}",
                    obj=prompt_template_dict,
                    error_msg=f"Could update variant {variant} of prompt template {template_id} for project {self._project_id}",
                    network_errors={
                        400: "Please check the prompt_variant or the arguments.",
                        403: f"Could not access project {self._project_id}. Please check the API key.",
                    },
                )
            return prompt_template_dict["variant"], prompt_template_dict["template"]

    def _sync_stash(
        self,
        template_id: str,
        variant: str,
        prompt: str,
        output: str,
        inputs: str = None,
        other: dict = None,
    ):
        prompt_embedding, err = get_embedding(prompt)
        if err:
            warn("Error generating embedding for prompt in child thread.")
            raise AIHeroException(err)
        output_embedding, err = get_embedding(output)
        if err:
            warn("Error generating embedding for output in child thread.")
            raise AIHeroException(err)

        stash_obj = {
            "prompt": prompt,
            "output": output,
            "prompt_embedding": prompt_embedding,
            "output_embedding": output_embedding,
        }

        if inputs:
            inputs_embedding, err = get_embedding(inputs)
            if err:
                warn("Error generating embedding for output in child thread.")
                raise AIHeroException(err)
            stash_obj["inputs"] = inputs
            stash_obj["inputs_embedding"] = inputs_embedding

        if other is None:
            other = {}
        other.update({"embedding_model": "text-embedding-ada-002"})
        stash_obj["other"] = other

        self._client.post(
            f"/tools/promptstash/projects/{self._project_id}/prompt_templates/{template_id}/variants/{variant}/stash",
            obj=stash_obj,
            timeout=30,
        )
        print("Prompt stashed.")

    def stash(
        self,
        variant: str,
        prompt: str,
        output: str,
        inputs: str = None,
        other: dict = None,
    ):
        if has_key():
            Thread(
                target=self._sync_stash,
                args=(variant, prompt, output, inputs, other),
            ).start()
        else:
            raise AIHeroException("No OPENAI_API_KEY in env variables.")



import httpx
from .exceptions import AIHeroException
import traceback


PRODUCTION_URL = "https://app.aihero.studio"
SANDBOX_URL = "https://sandbox.aihero.studio"


class Client:
    """Abstraction for http operations"""

    _base_url: str = None
    _authorization: str = None

    def __init__(self, bearer_token: str, server_url: str = PRODUCTION_URL):
        self._bearer_token = bearer_token
        self._authorization = f"Bearer {self._bearer_token}"
        self._base_url = server_url
        print(f"Connecting to {self._base_url}")
        if self._base_url.endswith("/"):
            self._base_url = self._base_url[:-1]
        self._base_url = f"{self._base_url}/api/v1"

    def _get_headers(self) -> dict:
        headers = {
            "Content-Type": "application/json",
            "Authorization": self._authorization,
        }
        return headers

    def get(
        self,
        path: str,
        error_msg: str = None,
        network_errors: dict = None,
        timeout: int = 30,
    ) -> dict:
        with httpx.Client(base_url=self._base_url, timeout=timeout) as client:
            response = client.get(path, headers=self._get_headers())
            try:
                response.raise_for_status()
            except httpx.HTTPStatusError as exc:
                msg = ""
                if network_errors and exc.response.status_code in network_errors:
                    raise AIHeroException(
                        f"{error_msg}: {network_errors[exc.response.status_code]} - {exc.response.text}",
                        status_code=exc.response.status_code,
                    ) from exc
                elif exc.response.status_code:
                    raise AIHeroException(
                        error_msg + "-" + exc.response.text,
                        status_code=exc.response.status_code,
                    ) from exc
                else:
                    msg = error_msg
                raise AIHeroException(msg) from exc
            return response.json()

    def post(
        self,
        path: str,
        obj: dict,
        error_msg: str = "Error",
        network_errors: dict = None,
        timeout: int = 30,
    ) -> dict:
        with httpx.Client(base_url=self._base_url, timeout=timeout) as client:
            response = client.post(path, json=obj, headers=self._get_headers())
        try:
            response.raise_for_status()
        except httpx.HTTPStatusError as exc:
            msg = ""
            if network_errors and exc.response.status_code in network_errors:
                raise AIHeroException(
                    f"{error_msg}: {network_errors[exc.response.status_code]} - {exc.response.text}",
                    status_code=exc.response.status_code,
                ) from exc
            elif exc.response.status_code:
                raise AIHeroException(
                    error_msg + "-" + exc.response.text,
                    status_code=exc.response.status_code,
                ) from exc
            else:
                msg = error_msg
            raise AIHeroException(msg) from exc
        return response.json()



from .client import Client
from .promptstash import PromptStash


class Project:
    """Main sync client class to talk to AI Hero"""

    def __init__(
        self,
        project_id: str,
        api_key: str,
        server_url: str = "https://app.aihero.studio",
    ):
        self._project_id = project_id
        self._client = Client(bearer_token=api_key, server_url=server_url)
        self.get()  # check

    def get(self, verbose: bool = False):
        project_dict = self._client.get(
            f"/workspace/projects/{self._project_id}",
            error_msg=f"Could fetch project details for project {self._project_id}",
            network_errors={
                400: "Please check the project_id.",
                403: f"Could not get project {self._project_id}. Please check the API key.",
            },
        )
        if verbose:
            print(f"Project id:\t{project_dict['project_id']}")
            print(f"Project name:\t{project_dict['name']}")
            print(f"Project desc:\t{project_dict['description']}")
            print(f"Created on:\t{project_dict['created_at']}")

    def promptstash(self):
        return PromptStash(self._project_id, self._client)



import os

import httpx

BASE_URL = "https://api.openai.com/v1/"
COMPLETIONS = "completions"
MODEL = "text-davinci-003"
EMBEDDINGS = "embeddings"
EMBEDDINGS_MODEL = "text-embedding-ada-002"
OPENAI_API_KEY = os.environ.get("OPENAI_API_KEY")


def has_key():
    return OPENAI_API_KEY is not None


def complete_with_gpt3(row: dict, template: any, api_key: str = OPENAI_API_KEY):
    prompt = template.render(**row)
    try:
        with httpx.Client(
            base_url=BASE_URL,
            headers={
                "Content-Type": "application/json",
                "Authorization": f"Bearer {api_key}",
            },
        ) as client:
            resp = client.post(
                COMPLETIONS,
                json={"model": MODEL, "prompt": prompt, "max_tokens": 512},
                timeout=30,
            )
            resp.raise_for_status()
            prediction_object = resp.json()
            if "errors" in prediction_object:
                return None, prediction_object["errors"][0]["text"].strip()
            else:
                completion = prediction_object["choices"][0]["text"].strip()
                print(f"Prompt:{prompt}\nCompletion:{completion}")
                return completion, None
    except httpx.HTTPError as e:
        return None, str(e)


def get_embedding(text: str, api_key: str = OPENAI_API_KEY):
    try:
        with httpx.Client(
            base_url=BASE_URL,
            headers={
                "Content-Type": "application/json",
                "Authorization": f"Bearer {api_key}",
            },
        ) as client:
            resp = client.post(
                EMBEDDINGS,
                json={"model": EMBEDDINGS_MODEL, "input": text},
                timeout=30,
            )
            resp.raise_for_status()
            prediction_object = resp.json()
            if "errors" in prediction_object:
                return None, prediction_object["errors"][0]["text"].strip()
            else:
                embedding = prediction_object["data"][0]["embedding"]
                return embedding, None
    except httpx.HTTPError as http_error:
        return None, str(http_error)
